#include "lifenet.hpp"
#include "string.hpp"
#include "objecthash.hpp"

inline bool char_is_whitespace(char ch) {
  return (ch == ' ' || ch == '\n' || ch == '\r' || ch == '\t');
}

inline void link_parse_error() {
  throw (LifeNetLinkExceptionPtr)(new LifeNetLinkException)->
    init(__FILE__, __LINE__, "LifeNetLink.init", "could not parse link");
}

int lifenetlink_parse_lisp_str(int       link_version,
			       char*     lisp_str,     // input
			       bool*     is_temporal,  // outputs (valid objects or NULL)
			       real*     plausibility,
			       StringPtr tag0,
			       StringPtr tag1,
			       int*      id0,
			       int*      id1) {
  int i = 0;
  switch(link_version) {
  case LINK_VERSION_0_23: {
    int j;
    char quote_char;
    // format: "(=> 'tag0' 'tag1' 0.5)"
    //sscanf(lisp_str, " ( %s '%s' '%s' %f ) ", 
    
    // find begin paren
    while(char_is_whitespace(lisp_str[i])) { i++; }
    if(lisp_str[i] != '(') {
      printf("parse error 0 '%c'\n", lisp_str[i]);
      link_parse_error();
    }
    i++;
    // read temporal marker symbol
    while(char_is_whitespace(lisp_str[i])) { i++; };
    if(lisp_str[i] == '-') {
      if(lisp_str[i+1] == '>' && char_is_whitespace(lisp_str[i+2])) {
	*is_temporal = false;
      } else {
	printf("parse error 1\n");
	link_parse_error();
      }
    } else if(lisp_str[i] == '=') {
      if(lisp_str[i+1] == '>' && char_is_whitespace(lisp_str[i+2])) {
	*is_temporal = true;
      } else {
	printf("parse error 2\n");
	link_parse_error();
      }
    } else {
      printf("parse error 3 '%c'\n", lisp_str[i]);
      link_parse_error();
    }
    i += 3;
    
    // read tag0 string
    while(char_is_whitespace(lisp_str[i])) { i++; };
    if(lisp_str[i] != '\'' && lisp_str[i] != '"') {
      printf("parse error 4\n");
      link_parse_error();
    }
    quote_char = lisp_str[i++];
    j = i;
    while(lisp_str[i] != quote_char) {
      if(lisp_str[i] == 0) {
	link_parse_error();
      }
      i++;
    }
    tag0->set_str(lisp_str+j, i-j);
    i++;
    
    // read tag1 string
    while(char_is_whitespace(lisp_str[i])) { i++; };
    if(lisp_str[i] != '\'' && lisp_str[i] != '"') {
      printf("parse error 5\n");
      link_parse_error();
    }
    quote_char = lisp_str[i++];
    j = i;
    while(lisp_str[i] != quote_char) {
      if(lisp_str[i] == 0) {
	link_parse_error();
      }
      i++;
    }
    tag1->set_str(lisp_str+j, i-j);
    i++;
    
    *plausibility = atof(lisp_str+i);
    
    // find end paren
    while(lisp_str[i] != ')' && lisp_str[i] != 0) {
      i++;
    }
    if(lisp_str[i] != ')') {
      printf("Link Parse Error: Couldn't find end parenthesis.\n");
      link_parse_error();
    }
    i++;
    while(char_is_whitespace(lisp_str[i])) { i++; }
  } break;
  case LINK_VERSION_0_22: {
    bool is_link = false;
    while(lisp_str[i] != '(') {
      if(lisp_str[i] == 0) {
	printf("Link parse error: couldn't find begin parenthesis.\n");
	link_parse_error();
      }
      i++;
    }
    if(strncmp(lisp_str+i, "(=> ", 4) == 0) {
      is_link     = true;
      if(is_temporal != NULL) {
	*is_temporal = true;
      }
    } else if(strncmp(lisp_str+i, "(-> ", 4) == 0) {
      is_link     = true;
      if(is_temporal != NULL) {
	*is_temporal = false;
      }
    }
    if(!is_link) {
      printf("Link parse error: couldn't recognize temporal symbol.\n");
      link_parse_error();
    }
    if(plausibility != NULL) {
      *plausibility = atof(lisp_str+4);
    }
    char tag_str[1024];
    i += 11;
    int j;
    for(j = 0; lisp_str[i+j] != '"' && lisp_str[i+j] != 0; j++);
    if(lisp_str[i+j] == 0) {
      printf("Link parse error: couldn't find quote 1.\n");
      link_parse_error();
    }
    if(tag0 != NULL) {
      memcpy(tag_str, lisp_str+i, j);
      tag_str[j] = 0;
      tag0->set_str(tag_str);
    }
    i += j+3;
    for(j = 0; lisp_str[i+j] != '"' && lisp_str[i+j] != 0; j++);
    if(lisp_str[i+j] == 0) {
      printf("Link parse error: couldn't find quote 2.\n");
      link_parse_error();
    }
    if(tag1 != NULL) {
      memcpy(tag_str, lisp_str+i, j);
      tag_str[j] = 0;
      tag1->set_str(tag_str);
    }
    i += j+2;
    if(id0 != NULL) {
      *id0 = atoi(lisp_str+i);
    }
    for(j = 0; lisp_str[i+j] != ' ' && lisp_str[i+j] != 0; j++);
    if(lisp_str[i+j] == 0) {
      printf("Link parse error: couldn't find plausibility.\n");
      link_parse_error();
    }
    i += j+1;
    if(id1 != NULL) {
      *id1 = atoi(lisp_str+i);
    }    
    while(lisp_str[i] != ')' && lisp_str[i] != 0) { i++; }
    i++;
    while(char_is_whitespace(lisp_str[i])) { i++; }
  } break;
  }
  return i;
}

void lifenetlink_fwrite(FILE*     fptr,
			bool      is_temporal,
			real      plausibility,
			StringPtr tag0,
			StringPtr tag1,
			int       id0,
			int       id1) {
  fprintf(fptr, "(%s %.3f \"%s\" \"%s\" %d %d)",
	  is_temporal ? "=>":"->", plausibility, tag0->get_str(), tag1->get_str(), id0, id1);  
}

LifeNetPtr LifeNet::init() {
  TMRF::init();
  var_hash = (new ObjectHash)->init();
  return this;
}

ObjectPtr LifeNet::_destroy() {
  TMRF::_destroy();
  delete var_hash->destroy();
  return this;
}

StringPtr LifeNet::add_tag(StringPtr tag) {
  StringPtr added_tag = tag;
  if(!var_hash->contains(tag)) {
    added_tag = (new String)->init(tag);
    var_hash->set_key(added_tag);
    TMRF::add_var(added_tag);
  } else {
    added_tag = (StringPtr)var_hash->lookup(tag);
  }
  return added_tag;
}

StringPtr LifeNet::lookup_tag(StringPtr tag) {
  return (StringPtr)var_hash->lookup(tag);
}

void LifeNet::add_correlation_factor(pointer n0, pointer n1, real correlation) {
  FactorPtr        factor     = (new Factor)->init();
  DataPtr          state      = (new Data)->init();
  DataValueBoolPtr value_bool = (new DataValueBool)->init(false);
  factor->add_var(n0);
  factor->add_var(n1);
  
  state->add_value(0, n0, value_bool);
  state->add_value(0, n1, value_bool);
  factor->set_state_value(state, 0, 1.0);
  
  ((DataValueBoolPtr)state->get_value(0, n1))->set_bool(true);
  factor->set_state_value(state, 0, 1.0);
  
  ((DataValueBoolPtr)state->get_value(0, n0))->set_bool(true);
  factor->set_state_value(state, 0, (1.0 + correlation) / 4);
  
  ((DataValueBoolPtr)state->get_value(0, n1))->set_bool(false);
  factor->set_state_value(state, 0, (1.0 - correlation) / 4);
  
  bool factor_exists = true;
  try {
    find_factor(factor);
  } catch(ExceptionPtr e) {
    factor_exists = false;
  }
  if(!factor_exists) {
    add_factor(get_nonexistent_factor(), factor);
  } else {
    throw (LifeNetExceptionPtr)(new LifeNetException)->
      init(__FILE__, __LINE__, "LifeNet.add_correlation_factor", "Factor already in LifeNet graph.");
  }
  
  delete value_bool->destroy();
  delete state->destroy();  
  delete factor->destroy();
}

void LifeNet::add_correlation_pair(StringPtr tag0, StringPtr tag1, bool is_temporal, real correlation) {
  tag0 = add_tag(tag0);
  tag1 = add_tag(tag1);
  if(is_temporal) {
    add_correlation_factor(get_var_n0(tag0), get_var_n1(tag1), correlation);
  } else {
    add_correlation_factor(get_var_n0(tag0), get_var_n0(tag1), correlation);
  }
}

bool LifeNet::scan_links(int link_version, char* text) {
  bool success = true;
  int text_index = 0;
  printf("Parsing LifeNet links\n");
  StringPtr line = (new String)->init("");
  StringPtr tag0 = (new String)->init("");
  StringPtr tag1 = (new String)->init("");
  int temporal_link_count = 0;
  int prior_link_count = 0;
  int duplicate_factors_ignored = 0;
  while(success && text[text_index] != 0) {
    try {
      bool      is_temporal;
      real      plausibility;
      int       id0;
      int       id1;
      text_index +=
	lifenetlink_parse_lisp_str(link_version,
				   text + text_index, // input
				   &is_temporal,    // outputs (valid objects or NULL)
				   &plausibility,
				   tag0,
				   tag1,
				   &id0,
				   &id1);
      try {
	add_correlation_pair(tag0, tag1, is_temporal, plausibility);
	if(is_temporal) {
	  temporal_link_count ++;
	} else {
	  prior_link_count ++;
	}
	if(get_factor_num() % 10000 == 0) {
	  printf("loaded %u factors.\n", get_factor_num());
	}
      } catch(LifeNetExceptionPtr e) {
	duplicate_factors_ignored ++;
      }
    } catch(LifeNetLinkExceptionPtr e) {
      printf("Unsuccessful load.\n");
      success = false;
      delete e->destroy();	
    }
  }
  delete line->destroy();
  delete tag0->destroy();
  delete tag1->destroy();
  printf("loaded LifeNet with %d links (%d prior and %d temporal) connecting %d tags.\n",
	 get_factor_num(),
	 prior_link_count,
	 temporal_link_count,
	 var_hash->get_length());  
  if(duplicate_factors_ignored > 0) {
    printf("%d duplicate factors ignored (%g%% of total factors).\n",
	   duplicate_factors_ignored,
	   100.0 * (real)duplicate_factors_ignored / (real)(get_factor_num() +
							  duplicate_factors_ignored));
  }
  return success;
}

LifeNetPtr LifeNet::load(int link_version, char* filename) {
  FILE* fptr = fopen(filename, "rb");
  if(fptr == NULL) {
    throw (LifeNetExceptionPtr)(new LifeNetException)->
      init(__FILE__, __LINE__, "LifeNet.load", "couldn't open file for reading");
  }
  printf("Reading LifeNet file: %s\n", filename);
  fseek(fptr, 0, SEEK_END);
  dword file_size = ftell(fptr);
  char* file_text = new char[file_size];
  rewind(fptr);
  dword i;
  const dword block_size = 1024 * 1024;
  for(i = 0; i < file_size;) {
    i += fread(file_text+i, 1, (i + block_size >= file_size) ? (file_size - i) : block_size, fptr);
  }
  scan_links(link_version, file_text);
  delete[] file_text;
  return this;
}

int LifeNet::data_slice_add_phrase_prob(LifeNetDataPtr tdata,
					int            slice_index,
					StringPtr      phrase,
					real           prob) {
  int total_tags_affected = 0;
  DataValueBoolProbPtr value_prob = (new DataValueBoolProb)->init(prob);
  IteratorPtr i0 = new_var_iterator();
  while(i0->has_next()) {
    StringPtr var = (StringPtr)i0->next();
    if(var->contains_space_bounded(phrase)) {
      if(tdata->contains(slice_index, var)) {
	tdata->remove_value(slice_index, var);
      }
      tdata->add_value(slice_index, var, value_prob);
      //printf("found: \"%s\"\n", var->get_str());
      total_tags_affected ++;
    }
  }
  delete i0->destroy();
  delete value_prob->destroy();
  return total_tags_affected;
}

void LifeNet::data_slice_add_phrase_list_state(LifeNetDataPtr tdata,
					       int            slice_index,
					       ObjectListPtr  phrase_list,
					       bool           state) {
  DataValueBoolPtr value_bool = (new DataValueBool)->init(state);
  IteratorPtr i0 = new_var_iterator();
  while(i0->has_next()) {
    StringPtr var = (StringPtr)i0->next();
    ObjectIteratorPtr i1 = phrase_list->new_iterator();
    while(i1->has_next()) {
      StringPtr phrase = (StringPtr)i1->next();
      if(var->contains_space_bounded(phrase)) {
	//value_bool->set_bool(state);
	tdata->add_value(slice_index, var, value_bool);
      }
    }
    delete i1->destroy();
  }
  delete i0->destroy();
  delete value_bool->destroy();
}

LifeNetDataPtr LifeNet::new_known_data_from_phrase_list(ObjectListPtr phrase_list) {
  LifeNetDataPtr known_data = (LifeNetDataPtr)(new LifeNetData)->init(1);
  DataValueBoolPtr value_bool = (new DataValueBool)->init(true);
  IteratorPtr i0 = new_var_iterator();
  while(i0->has_next()) {
    StringPtr var = (StringPtr)i0->next();
    ObjectIteratorPtr i1 = phrase_list->new_iterator();
    while(i1->has_next()) {
      StringPtr phrase = (StringPtr)i1->next();
      if(var->contains_space_bounded(phrase)) {
	value_bool->set_bool(true);
	known_data->add_value(0, var, value_bool);
	break;
      }
    }
    delete i1->destroy();
  }
  delete i0->destroy();
  delete value_bool->destroy();
  return known_data;
}

LifeNetDataPtr LifeNet::new_phrase_list_inference_data(ObjectListPtr phrase_list,
						 real          changed_threshold) {
  LifeNetDataPtr known_data = new_known_data_from_phrase_list(phrase_list);
  printf("known data length: %d\n", known_data->get_length());
  LifeNetDataPtr inferred_data =
    (LifeNetDataPtr)(new LifeNetData)->init(known_data->get_slice_num());
  
  printf("inferring missing data\n");
  
  infer_missing_data(known_data, inferred_data, changed_threshold);
  printf("done inferring missing data\n");
  
  delete known_data->destroy();
  return inferred_data;
}

LifeNetDataPtr LifeNet::new_word_inference_data(char* query_words, real changed_threshold) {
  StringPtr user_str = (new String)->init(query_words);
  ObjectListPtr user_word_list = user_str->new_word_list();
  delete user_str->destroy();
  
  printf("user word list:");
  ObjectIteratorPtr iter = user_word_list->new_iterator();
  while(iter->has_next()) {
    StringPtr word = (StringPtr)iter->next();
    printf(" \"%s\"", word->get_str());
  }
  delete iter->destroy();    
  printf("\n");
  
  LifeNetDataPtr inferred_data = new_phrase_list_inference_data(user_word_list, changed_threshold);
  iter = user_word_list->new_iterator();
  while(iter->has_next()) {
    StringPtr word = (StringPtr)iter->next();
    delete word->destroy();
  }
  delete iter->destroy();    
  delete user_word_list->destroy();
  
  return inferred_data;
}

ObjectPtr LifeNet::_fwrite(FILE* fptr) {
  fprintf(fptr, "(lifenet");
  IteratorPtr iter = new_var_iterator();
  while(iter->has_next()) {
    StringPtr node = (StringPtr)iter->next();
    fprintf(fptr, "\n  (\"%s\")", node->get_str());
  }
  delete iter->destroy();
  fprintf(fptr, ")\n");
  return this;
}

ObjectPtr LifeNet::_print() {
  return _fwrite(stdout);
}

