#include "lifenet.hpp"
#include <export.h>

#define LIFENET_INFERENCE_ACCURACY 1e-2

MutexPtr lifenet_mutex = NULL;

void lifenet_mutex_lock() {
  if(lifenet_mutex == NULL) {
    lifenet_mutex = (new Mutex)->init();
  }
  lifenet_mutex->sure_lock();  
}

void lifenet_mutex_unlock() {
  lifenet_mutex->unlock();  
}

bool     lifenet_initialized = false;
HashPtr  lifenet_id_hash     = NULL;
HashPtr  data_id_hash        = NULL;

/** accepts a lifenet filename for a file of the format of lifenet-0.22.lisp or a null string.    
 */
void assure_initialized() {
  if(lifenet_initialized) {
    return;
  }
  lifenet_mutex_lock();
  if(data_id_hash == NULL) {
    data_id_hash = (new Hash)->init();
  }
  if(lifenet_id_hash == NULL) {
    lifenet_id_hash = (new Hash)->init();
  }
  lifenet_initialized = true;
  lifenet_mutex_unlock();
}

LifeNetDataPtr _get_data(int id) {
  LifeNetDataPtr data;
  try {
    data = (LifeNetDataPtr)data_id_hash->lookup((dword)id);
  } catch(ExceptionPtr e) {
    printf("\nLifeNet Error: No data object associated with key (%u).\n", id);
    throw e;
  }
  return data;
}

int _add_data(LifeNetDataPtr data) {
  int id = data_id_hash->get_nonexistent_key();
  data_id_hash->add((dword)id, (pointer)data);
  return id;
}

int _new_data(int slice_num) {
  return _add_data((LifeNetDataPtr)(new LifeNetData)->init(slice_num));
}

LifeNetPtr _get_lifenet(int lifenet_id) {
  LifeNetPtr lifenet;
  try {
    lifenet = (LifeNetPtr)lifenet_id_hash->lookup((dword)lifenet_id);
  } catch(ExceptionPtr e) {
    printf("\nLifeNet Error: No lifenet object associated with key (%u).\n", lifenet_id);
    throw e;
  }
  return lifenet;
}

int _add_lifenet(LifeNetPtr lifenet) {
  int lifenet_id = lifenet_id_hash->get_nonexistent_key();
  lifenet_id_hash->add((dword)lifenet_id, (pointer)lifenet);
  return lifenet_id;
}

int _new_lifenet() {
  return _add_lifenet((new LifeNet)->init());
}

void _data_to_new_output_array(LifeNetDataPtr data,
			       int* out_argc0, int** out_argc1, char**** out_argv,
			       int* out_real_c0, int** out_real_c1, real*** out_real_vector) {
  *out_argc0       = data->get_slice_num();
  *out_argc1       = (int*)   malloc(sizeof(int)    * *out_argc0);
  *out_argv        = (char***)malloc(sizeof(char**) * *out_argc0);
  *out_real_c0     = data->get_slice_num();
  *out_real_c1     = (int*)  malloc(sizeof(int)   * *out_real_c0);
  *out_real_vector = (real**)malloc(sizeof(real*) * *out_real_c0);  
  int slice_index;
  for(slice_index = 0; slice_index < data->get_slice_num(); slice_index++) {
    (*out_argc1)[slice_index]       = data->get_slice_value_num(slice_index);
    (*out_argv)[slice_index]        = (char**)malloc(sizeof(char*) * (*out_argc1)[slice_index]);
    (*out_real_c1)[slice_index]     = data->get_slice_value_num(slice_index);
    (*out_real_vector)[slice_index] = (real*)malloc(sizeof(real) * (*out_real_c1)[slice_index]);
    int i = 0;
    IteratorPtr i2 = data->new_slice_node_iterator(slice_index);
    while(i2->has_next()) {
      HashNodePtr      hash_node = (HashNodePtr)i2->next();
      DataValueProbPtr value     = (DataValueProbPtr)hash_node->get_ptr();
      StringPtr        node      = (StringPtr)hash_node->get_key();
      //printf("(%u \"%s\" %g)\n", slice_index, str->get_str(), value->get_prob());
      (*out_argv)[slice_index][i] = (char*)malloc(sizeof(char) * (node->get_length()+1));
      strcpy((*out_argv)[slice_index][i], node->get_str());
      (*out_real_vector)[slice_index][i] = value->get_prob();
      i++;
    }
    delete i2->destroy();
  }
}

void _data_top_n_to_new_output_array(LifeNetDataPtr data, int top_num,
				     int* out_argc0, int** out_argc1, char**** out_argv,
				     int* out_real_c0, int** out_real_c1,
				     real*** out_real_vector) {
  *out_argc0       = data->get_slice_num();
  *out_argc1       = (int*)   malloc(sizeof(int)    * *out_argc0);
  *out_argv        = (char***)malloc(sizeof(char**) * *out_argc0);
  *out_real_c0     = data->get_slice_num();
  *out_real_c1     = (int*)  malloc(sizeof(int)   * *out_real_c0);
  *out_real_vector = (real**)malloc(sizeof(real*) * *out_real_c0);
  int slice_index;
  for(slice_index = 0; slice_index < data->get_slice_num(); slice_index++) {
    ObjectArrayPtr top_sorted       = data->new_slice_top_nodes_by_prob(slice_index, top_num);
    (*out_argc1)[slice_index]       = top_sorted->get_length();
    (*out_argv)[slice_index]        = (char**)malloc(sizeof(char*) * (*out_argc1)[slice_index]);
    (*out_real_c1)[slice_index]     = top_sorted->get_length();
    (*out_real_vector)[slice_index] = (real*)malloc(sizeof(real) * (*out_real_c1)[slice_index]);
    int i = 0;
    ObjectIteratorPtr i2 = top_sorted->new_iterator();
    while(i2->has_next()) {
      Hash2DNodePtr    hash2d_node = (Hash2DNodePtr)i2->next();
      DataValueProbPtr value       = (DataValueProbPtr)hash2d_node->get_ptr();
      StringPtr        node        = (StringPtr)hash2d_node->get_key()[1];
      printf("(%u \"%s\" %g)\n", slice_index, node->get_str(), value->get_prob());
      (*out_argv)[slice_index][i] = (char*)malloc(sizeof(char) * (node->get_length()+1));
      strcpy((*out_argv)[slice_index][i], node->get_str());
      (*out_real_vector)[slice_index][i] = value->get_prob();
      i++;
    }
    delete i2->destroy();
    delete top_sorted->destroy();
  }
}

EXPORT int lifenet_data_new(int slice_num) {
  int id = -1;
  assure_initialized();
  lifenet_mutex_lock();
  try {
    id = _new_data(slice_num);
  } catch(ExceptionPtr e) {
    log_printf("\nFunction Syntax: lifenet_data_new(int slice_num)\n");
    e->log_print();
  }
  lifenet_mutex_unlock(); 
  return id;
}

EXPORT void lifenet_data_delete(int id) {
  assure_initialized();
  lifenet_mutex_lock();
  try {
    delete _get_data(id)->destroy();
    data_id_hash->remove(id);
  } catch(ExceptionPtr e) {
    log_printf("\nFunction Syntax: lifenet_data_delete(int id)\n");
    e->log_print();
  }
  lifenet_mutex_unlock(); 
}

EXPORT int lifenet_data_add_slices(int data_id, int lifenet_id, char* slices) {
  bool success = false;
  assure_initialized();
  lifenet_mutex_lock();
  try {
    success = _get_data(data_id)->scan_slices(_get_lifenet(lifenet_id), slices);
  } catch(ExceptionPtr e) {
    log_printf("\nFunction Syntax: "
	       "lifenet_data_add_slices(int data_id, int lifenet_id, char* slices)\n");
    e->log_print();
  }
  lifenet_mutex_unlock();
  return success ? 0 : 1; // returns zero on success
}

EXPORT void lifenet_data_print(int id) {
  assure_initialized();
  lifenet_mutex_lock();
  try {
    FILE* fptr = stdout;
    LifeNetDataPtr data = _get_data(id);
    fprintf(fptr, "(lifenet-data");
    int slice_index;
    for(slice_index = 0; slice_index < data->get_slice_num(); slice_index++) {
      fprintf(fptr, "\n  (slice %d", slice_index);
      HashIteratorPtr hash_iter = (HashIteratorPtr)data->new_slice_node_iterator(slice_index);
      while(hash_iter->has_next()) {
	HashNodePtr    hash_node = (HashNodePtr)hash_iter->next();
	StringPtr      node      = (StringPtr)hash_node->get_key();
	DataValuePtr   value     = (DataValuePtr)hash_node->get_ptr();
	fprintf(fptr, "\n    (\"%s\" ", node->get_str());
	value->fwrite(fptr);
	fprintf(fptr, ")");
      }
      delete hash_iter->destroy();
      fprintf(fptr, ")");
    }
    fprintf(fptr, ")\n");
  } catch(ExceptionPtr e) {
    log_printf("\nFunction Syntax: lifenet_data_print(int id)\n");
    e->log_print();
  }
  lifenet_mutex_unlock(); 
}

EXPORT void lifenet_data_add_phrases(int lifenet_id, int data_id,
				     int state, int argc0, int* argc1, char*** argv) {
  assure_initialized();
  lifenet_mutex_lock();
  try {
    LifeNetPtr lifenet = _get_lifenet(lifenet_id);
    LifeNetDataPtr data = _get_data(data_id);
    ObjectListPtr phrase_list;
    StringPtr str;
    int i, j;
    for(i = 0; i < argc0; i++) { // slice iterator
      phrase_list = (new ObjectList)->init();
      for(j = 0; j < argc1[i]; j++) {
	str = (new String)->init(argv[i][j]);
	phrase_list->add(str);
      }
      lifenet->data_slice_add_phrase_list_state(data, i, phrase_list, state != 0);
      ObjectIteratorPtr iter = phrase_list->new_iterator();
      while(iter->has_next()) {
	str = (StringPtr)iter->next();
	delete str->destroy();
      }
      delete phrase_list->destroy();
    }
  } catch(ExceptionPtr e) {
    log_printf("\nFunction Syntax: lifenet_data_add_phrases(int id, int state, List of List of Strings)\n");
    e->log_print();
  }
  lifenet_mutex_unlock();   
}

EXPORT void lifenet_data_get_contents(int id,
				      int* out_argc0, int** out_argc1, char**** out_argv,
				      int* out_real_c0, int** out_real_c1, real*** out_real_vector)
{
  assure_initialized();
  lifenet_mutex_lock();
  try {
    _data_to_new_output_array(_get_data(id),
			      out_argc0, out_argc1, out_argv,
			      out_real_c0, out_real_c1, out_real_vector);
  } catch(ExceptionPtr e) {
    log_printf("\nFunction Syntax: lifenet_data_get_contents(int id)\n");
    e->log_print();
    *out_argc0   = 0;
    *out_real_c0 = 0;
  }
  lifenet_mutex_unlock(); 
}

EXPORT void lifenet_data_get_top_n_contents(int id, int top_num,
					    int* out_argc0, int** out_argc1, char**** out_argv,
					    int* out_real_c0, int** out_real_c1,
					    real*** out_real_vector) {
  assure_initialized();
  lifenet_mutex_lock();
  try {
    _data_top_n_to_new_output_array(_get_data(id), top_num,
				    out_argc0, out_argc1, out_argv,
				    out_real_c0, out_real_c1, out_real_vector);
  } catch(ExceptionPtr e) {
    log_printf("\nFunction Syntax: lifenet_data_get_top_n_contents(int id, int top_num)\n");
    e->log_print();
    *out_argc0   = 0;
    *out_real_c0 = 0;
  }
  lifenet_mutex_unlock(); 
}

EXPORT void lifenet_data_subtract(int a_id, int b_id) {
  assure_initialized();
  lifenet_mutex_lock();
  try {
    LifeNetDataPtr data_a = _get_data(a_id);
    LifeNetDataPtr data_b = _get_data(b_id);
    data_a->subtract_prob_data(data_b);
  } catch(ExceptionPtr e) {
    log_printf("\nFunction Syntax: lifenet_data_subtract(int a_id, int b_id)\n");
    e->log_print();
  }
  lifenet_mutex_unlock(); 
}

EXPORT void lifenet_data_copy(int a_id, int b_id) {
  assure_initialized();
  lifenet_mutex_lock();
  try {
    LifeNetDataPtr data_a = _get_data(a_id);
    LifeNetDataPtr data_b = _get_data(b_id);
    data_a->copy(data_b);
  } catch(ExceptionPtr e) {
    log_printf("\nFunction Syntax: lifenet_data_copy(int a_id, int b_id)\n");
    e->log_print();
  }
  lifenet_mutex_unlock(); 
}

EXPORT int lifenet_new() {
  int lifenet_id = -1;
  assure_initialized();
  lifenet_mutex_lock();
  try {
    lifenet_id = _new_lifenet();
  } catch(ExceptionPtr e) {
    log_printf("\nFunction Syntax: lifenet_new()\n");
    e->log_print();
  }
  lifenet_mutex_unlock(); 
  return lifenet_id;
}

EXPORT void lifenet_load(int lifenet_id, char* filename) {
  assure_initialized();
  lifenet_mutex_lock();
  try {
    LifeNetPtr lifenet = _get_lifenet(lifenet_id);
    if(filename != NULL && filename[0] != 0) {
      lifenet->load(LINK_VERSION_0_22, filename);
      printf("Calculating default data.\n");
      lifenet->recalculate_default_data(LIFENET_INFERENCE_ACCURACY);
    }
  } catch(ExceptionPtr e) {
    log_printf("\nFunction Syntax: lifenet_load(int lifenet_id, Filename String)\n");
    e->log_print();
  }
  lifenet_mutex_unlock(); 
}

EXPORT void lifenet_delete(int lifenet_id) {
  assure_initialized();
  lifenet_mutex_lock();
  try {
    delete _get_lifenet(lifenet_id)->destroy();
    lifenet_id_hash->remove(lifenet_id);
  } catch(ExceptionPtr e) {
    log_printf("\nFunction Syntax: lifenet_destroy(int lifenet_id)\n");
    e->log_print();
  }
  lifenet_mutex_unlock();
}

EXPORT void lifenet_infer(int lifenet_id, int known_data_id, int inferred_data_id) {
  assure_initialized();
  lifenet_mutex_lock();
  try {
    _get_lifenet(lifenet_id)->infer_missing_data(_get_data(known_data_id),
						 _get_data(inferred_data_id),
						 LIFENET_INFERENCE_ACCURACY);
  } catch(ExceptionPtr e) {
    log_printf("\nFunction Syntax: lifenet_infer(int lifenet_id, int known_data_id, int inferred_data_id)\n");
    e->log_print();
  }
  lifenet_mutex_unlock(); 
}

EXPORT void lifenet_fwrite(int lifenet_id, FILE* fptr) {
  assure_initialized();
  lifenet_mutex_lock();
  try {
    _get_lifenet(lifenet_id)->fwrite(fptr);
  } catch(ExceptionPtr e) {
    log_printf("\nFunction Syntax: lifenet_fwrite(int lifenet_id, FILE* fptr)\n");
    e->log_print();
  }
  lifenet_mutex_unlock(); 
}

EXPORT void lifenet_print(int lifenet_id) {
  assure_initialized();
  lifenet_mutex_lock();
  try {
    _get_lifenet(lifenet_id)->print();
  } catch(ExceptionPtr e) {
    log_printf("\nFunction Syntax: lifenet_print(int lifenet_id)\n");
    e->log_print();
  }
  lifenet_mutex_unlock(); 
}

EXPORT void lifenet_add_edge(int lifenet_id, char* tag0_str, char* tag1_str, int is_temporal,
			     real plausibility) {
  assure_initialized();
  lifenet_mutex_lock();
  try {
    StringPtr tag0 = (new String)->init(tag0_str);
    StringPtr tag1 = (new String)->init(tag1_str);
    _get_lifenet(lifenet_id)->add_correlation_pair(tag0, tag1, is_temporal!=0, plausibility);
    delete tag0->destroy();
    delete tag1->destroy();
  } catch(ExceptionPtr e) {
    log_printf("\nFunction Syntax: lifenet_print(int lifenet_id)\n");
    e->log_print();
  }
  lifenet_mutex_unlock();  
}

EXPORT int lifenet_add_edges(int lifenet_id, char* edges) {
  bool success = false;
  assure_initialized();
  lifenet_mutex_lock();
  try {
    success = _get_lifenet(lifenet_id)->scan_links(LINK_VERSION_0_23, edges);
  } catch(ExceptionPtr e) {
    log_printf("\nFunction Syntax: lifenet_add_edges(int lifenet_id, char* edges)\n");
    e->log_print();
  }
  lifenet_mutex_unlock();  
  return success ? 0 : 1; // zero on success
}

void fwrite_lifenet_data(FILE* fptr, LifeNetDataPtr data) {
  fprintf(fptr, "(lifenet-data");
  int slice_index;
  for(slice_index = 0; slice_index < data->get_slice_num(); slice_index++) {
    fprintf(fptr, "\n  (slice %d");
    HashIteratorPtr hash_iter = (HashIteratorPtr)data->new_slice_node_iterator(slice_index);
    while(hash_iter->has_next()) {
      HashNodePtr    hash_node = (HashNodePtr)hash_iter->next();
      StringPtr      node      = (StringPtr)hash_node->get_key();
      DataValuePtr   value     = (DataValuePtr)hash_node->get_ptr();
      fprintf(fptr, "\n    (\"%s\"", (int)node->get_str());
      value->fwrite(fptr);
      fprintf(fptr, ")");
    }
    delete hash_iter->destroy();
    fprintf(fptr, ")");
  }
  fprintf(fptr, ")\n");
}

int main(int argc, char** argv) {
  try {    
    if(argc < 3) {
      printf("lifenet <lifenet-0.22.lisp filename> -i \"list of words\"\n");
    } else {
      //printf("loading lifenet\n");
      LifeNetPtr lifenet = (new LifeNet)->init()->load(LINK_VERSION_0_22, argv[1]);
      
      LifeNetDataPtr inferred_data = lifenet->new_word_inference_data(argv[2], 0.01);
      
      fwrite_lifenet_data(stdout, inferred_data);
      
      delete inferred_data->destroy();
      
      //printf("destroying lifenet\n");
      delete lifenet->destroy();
      //printf("destroyed lifenet\n");
    }
  } catch(ExceptionPtr e) {
    e->handle_unexpected_internal();
  }
  return 0;
}

