(require :aserve)
(require :pxml)
(require :mysql)

(defpackage :openmind
  (:use :common-lisp :net.aserve :net.html.generator :net.xml.parser :dbi.mysql :think.utils))

(in-package :openmind)


;; External parameters
(defvar cl-user::*natlang-data* (merge-pathnames "apps/data/" cl-user::*think-directory-root*))
(defvar cl-user::*lifenet-root* (merge-pathnames "apps/lifenet/" cl-user::*think-directory-root*))
(defvar cl-user::*storynet-root* (merge-pathnames "apps/storynet/" cl-user::*think-directory-root*))

(defun init-lifenet ()
  ;; Create the db connection prior to launching the server
  (kb-connect))

;;;~~~~~~~~~~~~~~~~  Parse XML  ~~~~~~~~~~~~~~~~

(defparameter *xml*
   "<op><p id=\"10\" time=\"0\" truth=\"0\" /><p id=\"20\" time=\"0\" truth=\"0\" /></op>")

(defun get-rule-updates (xml)
   (let ((xml-list (parse-xml xml)))
    (loop for elt in (cdar xml-list)
            for id = (read-from-string (third (car elt)))
            for time = (read-from-string (fifth (car elt)))
            for truth = (read-from-string (seventh (car elt)))
            collect (list id time truth))))

(defun get-props-from-client (xml)
   (let ((xml-list (parse-xml xml)))
    (loop for elt in (cdar xml-list)
            for id = (read-from-string (third (car elt)))
            for time = 0
            for truth = 1
            collect (list id time truth))))

;;;~~~~~~~~~~~~~~~~  Natural Language functions  ~~~~~~~~~~~~~~~~

;; stopwords

(defvar stop-words (make-hash-table :test #'equal))

(defun load-stopwords ()
  (format t "Loading stopwords...")
  (format t "stopword filename: ~A~%" (merge-pathnames "stopwords.txt" cl-user::*natlang-data*))
  (with-open-file (f (merge-pathnames "stopwords.txt" cl-user::*natlang-data* ))
    (loop for line = (read-line f nil) while line do
      (setf (gethash line stop-words) t)))
  (format t "done.~%"))

(defun stopword? (word)
  (gethash word stop-words))

;; dictionary for lemmatising

(defvar lexeme-dict (make-hash-table :test #'equal))

(defun load-dictionary ()
  (format t "Loading dictionary...")
  (with-open-file (f (pathname (merge-pathnames cl-user::*natlang-data* (make-pathname :name "morph-truncated-fixed" :type "txt"))))
;;    (let ((extract (compile-regexp "^\\(\\B*\\)\\b*\\(\\B*\\)\\b*")))
    (let ((extract (pregex:create-scanner "^\\(\\B*\\)\\b*\\(\\B*\\)\\b*" 
					  :single-line-mode t :destructive t)))
      (loop for line = (read-line f nil) while line do
            (let* ((results (multiple-value-list
;;                             (match-regexp extract line :return :string)))
			     (pregex:all-matches-as-strings extract line)))
                   (word (string-downcase (third results)))
                   (lexeme (string-downcase (fourth results))))
              (setf (gethash word lexeme-dict) lexeme)))))
  (format t "done.~%"))

(defun lemmatise (word)
  (let ((result (gethash (string-downcase word) lexeme-dict)))
    (if result result word)))

;; string operations

(defparameter split-expr (pregex:create-scanner " "))

(defun lowered-words-in-string (str)
  (pregex:split split-expr (string-downcase str)))

(defun simplify (str)
  (reduce (lambda (x y) (format nil "~A ~A" x y))
    (mapcar #'lemmatise (pregex:split split-expr (string-downcase str)))))

(format t "Loading stopwords and dictionary.~%")
(load-stopwords)
(load-dictionary)
(format t "Done.~%")

;;;~~~~~~~~~~~~~~~~  Situations  ~~~~~~~~~~~~~~~~

;; load situations

(defparameter *situations* (make-hash-table))
(defparameter *indexed-situations* (make-hash-table :test #'equal))

(defun get-situations (word)
  (hash-get *indexed-situations* (lemmatise word)))

(defun get-situations-text (word)
  (mapcar (lambda (x) (cons x (hash-get *situations* x))) (get-situations word)))

(defun get-situation-n (n)
  (gethash n *situations*))

(defun get-best-sits (expr &optional (n 5))
   ;; return 
  (let ((candidates nil)
        (expr-words (remove-duplicates (lowered-words-in-string expr) :test #'equal))
        (results nil))
    (loop for word in expr-words do
      (unless (stopword? word)
        (setf candidates (append candidates (get-situations-text word)))))
    (loop for candidate in candidates do
      (let ((score 0))
        (loop for w1 in (remove-duplicates (lowered-words-in-string (cdr candidate)) :test #'equal) do
          (loop for w2 in expr-words do
            (unless (or (stopword? w1) (stopword? w2))
              (if (string-equal (lemmatise w1) (lemmatise w2))
                (incf score 2)
                (decf score)))))
        (setf results (push (list score (car candidate) (cdr candidate)) results))))
    (subseq (sort results #'> :key #'car) 0 n)))

(defun load-situations ()
  (format t "Loading Lifenet...")
  (with-open-file (f (merge-pathnames cl-user::*lifenet-root* (make-pathname :name "lifenet-sits" :type "lisp")))
    (loop for i upfrom 0
          for sit = (read f nil) while sit do
          (when (= 0 (mod i 10000))
             (format t "~A...~%" (/ i 10000)))
          (let ((sit-id (first sit))
                 (sit-text (second sit)))
            (unless (has-key? *situations* sit-id)
              (hash-put *situations* sit-id sit-text))))))

(defun index-situations ()
  (format t "Indexing situations...")
  (maphash (lambda (i sit)
                   (if (= 0 (mod i 1000)) (format t "~A...~%" (/ i 1000)))
                   (loop for word in (lowered-words-in-string sit) do
                           (hash-push *indexed-situations* (lemmatise word) i)))
      *situations*)
  (format t "done.~%"))

(load-situations)
(index-situations)

;;;~~~~~~~~~~~~~~~~  Knowledge Base API  ~~~~~~~~~~~~~~~~

(defun kb-connect ()
   (connect :database "omlife" :user "root" :password "omsql"))

(defun kb-sql (str)
  ;; Wrapper for sql operations. Needed because Mysql disconnects automatically after 8 hours.
  (catch 'trap-error
     (handler-bind ((dbi.mysql::mysql-error
                           #'(lambda (err) (declare (ignore err)) (kb-connect) (throw 'trap-error (sql str)))))
             (sql str))))

(defun kb-link-member? (p1 p2 temporal?)
  ;; Returns T iff Lifenet KB has link between proposition ids P1 and P2. Looks in
  ;; the temporal table if TEMPORAL? is true, else looks in the atemporal table.
  (if temporal?
    (if (kb-sql (format nil "select 1 from temporal where orig = ~A and dest = ~A" p1 p2)) t)
    (if (kb-sql (format nil "select 1 from atemporal where orig = ~A and dest = ~A" p1 p2)) t)))

(defun kb-search-props (query)
  ;; Returns list of proposition ids whose text matches QUERY.
   (let ((results (kb-sql (format nil "select prob, id from props where match (textkeys) against ('~A')" query))))
    (if results
       (mapcar #'second (sort results (lambda (x y) (> (first x) (first y))))))))

(defun kb-get-prop-by-id (id)
  (let ((results (kb-sql (format nil "select text from props where id = ~A" id))))
    (if results
      (caar results))))

(defun kb-get-links (propid t1 t2)
  ;; Given proposition id PROPID true at T1, return list of propositions it links to at T2.
  (let ((results
         (cond ((= t1 t2)
                  (kb-sql (format nil "select prob, dest from atemporal where orig = ~A" propid)))
                 ((= t1 0)
                  (kb-sql (format nil "select prob, dest from temporal where orig = ~A" propid)))
                 ((= t1 1)
                  (kb-sql (format nil "select prob, orig from temporal where dest = ~A" propid))))))
   (sort results (lambda (x y) (> (first x) (first y))))))

(defun kb-get-links-loose (propid t1 t2)
  ;; Given proposition id PROPID true at T1, return list of propositions it links to at T2, through
  ;; any intermediate props.
   (let* ((sit-text (get-situation-n propid))
           (relatedpropids (mapcar #'second (get-best-sits sit-text 3))))
    (let ((results nil))
     (loop for id in (cons propid relatedpropids)
             for result =
              (cond
               ((= t1 t2)
                (kb-sql (format nil "select prob, dest from atemporal where orig = ~A" id)))
               ((= t1 0)
                (kb-sql (format nil "select prob, dest from temporal where orig = ~A" id)))
               ((= t1 1)
                (kb-sql (format nil "select prob, orig from temporal where dest = ~A" id))))
             if result do
             (setf results (append result results)))
     (sort results (lambda (x y) (> (first x) (first y)))))))

(defun kb-suggest-completions (props)
  ;; Given a list of propositions supplied as (ID TIME TRUTH) elements, suggest what other
  ;; propositions might be true, returned as (ID TIME TRUTH) elements sorted from highest
  ;; to lowest probability.
  (let ((results nil))
   (loop for prop in props
          for id = (first prop)
          do
          (cond ((and (= (second prop) 0) (= (third prop) 1))
                   (setf results (append results (mapcar (lambda (x) (list (first x) (second x) 1 1)) (kb-get-links-loose id 0 1))))
                   (setf results (append results (mapcar (lambda (x) (list (first x) (second x) 0 1)) (kb-get-links-loose id 0 0)))))
                  ((and (= (second prop) 1) (= (third prop) 1))
                   (setf results (append results (mapcar (lambda (x) (list (first x) (second x) 0 1)) (kb-get-links-loose id 1 0))))
                   (setf results (append results (mapcar (lambda (x) (list (first x) (second x) 1 1)) (kb-get-links-loose id 1 1))))
                   (setf results (append results (mapcar (lambda (x) (list (first x) (second x) 2 1)) (kb-get-links-loose id 0 1)))))
                  ((and (= (second prop) 2) (= (third prop) 1))
                   (setf results (append results (mapcar (lambda (x) (list (first x) (second x) 1 1)) (kb-get-links-loose id 1 0))))
                   (setf results (append results (mapcar (lambda (x) (list (first x) (second x) 2 1)) (kb-get-links-loose id 1 1)))))))
   (remove-duplicates (mapcar #'rest (sort results (lambda (x y) (> (first x) (first y))))) :test #'equal)))

(defun kb-suggest-next (props)
  ;; Given a list of propositions supplied as (ID TIME TRUTH) elements, suggest what other
  ;; propositions might be true next, returned as (ID TIME TRUTH) elements sorted from highest
  ;; to lowest probability.
  (let ((results nil))
   (loop for prop in props
          for id = (first prop)
          do
           (setf results (append results (mapcar (lambda (x) (list (first x) (second x) 1 1)) (kb-get-links-loose id 0 1))))
           (setf results (append results (mapcar (lambda (x) (list (first x) (second x) 0 1)) (kb-get-links-loose id 0 0)))))
   (remove-duplicates (mapcar #'rest (sort results (lambda (x y) (> (first x) (first y))))) :test #'equal)))

(defun kb-get-prob (p1 p2 temporal?)
  ;; Returns probability of link between P1 and P2, or NIL if there is no link.
  ;; Looks in the temporal table if TEMPORAL? is true, else looks in the atemporal table.
  (if temporal?
    (kb-sql (format nil "select prob from temporal where orig = ~A and dest = ~A" p1 p2))
    (kb-sql (format nil "select prob from atemporal where orig = ~A and dest = ~A" p1 p2))))

(defun kb-put-link (p1 p2 temporal?)
  ;; Adds a link between propositions P1 and P2 of probability 1.0. If there is already a link updates
  ;; the probability to 1.0. Operates on the temporal table if TEMPORAL? is true, else looks in the
  ;; atemporal table. Does nothing if TEMPORAL? is false and P1 = P2.
  (cond ((and (= p1 p2) (not temporal?))
           nil)
          ((not temporal?)
           (let ((prob1 (kb-get-prob p1 p2 nil))
                  (prob2 (kb-get-prob p2 p1 nil)))
            (if prob1
             (kb-sql (format nil "update atemporal set prob = 1 where orig = ~A and dest = ~A" p1 p2))
             (kb-sql (format nil "insert into atemporal (orig, dest, prob) values (~A, ~A, ~A)" p1 p2 1.0)))
            (if prob2
             (kb-sql (format nil "update atemporal set prob = 1 where dest = ~A and orig = ~A" p1 p2))
             (kb-sql (format nil "insert into atemporal (dest, orig, prob) values (~A, ~A, ~A)" p1 p2 1.0)))))
          (t
           (let ((prob (kb-get-prob p1 p2 t)))
            (if prob
             (kb-sql (format nil "update temporal set prob = 1 where orig = ~A and dest = ~A" p1 p2))
             (kb-sql (format nil "insert into temporal (orig, dest, prob) values (~A, ~A, ~A)" p1 p2 1.0))))))
  (kb-sql (format nil "update props set prob = 1 where id = ~A" p1))
  (kb-sql (format nil "update props set prob = 1 where id = ~A" p2)))

(defun kb-update (updates)
  ;; Given a list of propositions supplied as (ID TIME TRUTH) elements, update lifenet to
  ;; include all pairwise links. If a link already exists, increase its probability of being true.
  (loop for prop1 in updates do
    (loop for prop2 in updates
          for id1 = (first prop1)
          for id2 = (first prop2)
          for time1 = (second prop1)
          for time2 = (second prop2)
          for truth1 = (third prop1)
          for truth2 = (third prop2)
          if (and (= 1 truth1) (= 1 truth2))
          do
          (cond ((< time1 time2) (kb-put-link id1 id2 t))
                   ((> time1 time2) (kb-put-link id2 id1 t))
                   ((= time1 time2) (kb-put-link id1 id2 nil))))))

(defun kb-get-random-goal ()
  (let ((results (kb-sql "select orig, dest from temporal where prob = 1 order by rand() limit 1")))
   (if results (first results))))


;;;~~~~~~~~~~~~~~~~  HTTP-XML Interface to LifeNet  ~~~~~~~~~~~~~~~~

(publish-file
 :path "/lifenet/LNT.swf"
 :file (merge-pathnames cl-user::*lifenet-root* "LNT.swf"))

(publish-file
 :path "/lifenet/flash_detection.swf"
 :file (merge-pathnames cl-user::*lifenet-root* (make-pathname :name "flash_detection" :type "swf")))

(publish :path "/lifenet/"
  :content-type "text/html"
  :function
  (lambda (req ent)
    (with-http-response (req ent)
      (with-http-body (req ent)
       (html (:princ 
"<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\">
<html xmlns=\"http://www.w3.org/1999/xhtml\" xml:lang=\"en\" lang=\"en\">
<head>
<meta http-equiv=\"refresh\" content=\"3;url=LNT_alternate.html\" />
<title>LNT</title>
</head>
<body>
<object classid=\"clsid:d27cdb6e-ae6d-11cf-96b8-444553540000\" codebase=\"http://download.macromedia.com/pub/shockwave/cabs/flash/swflash.cab#version=4,0,0,0\" width=\"80\" height=\"80\">
<param name=\"movie\" value=\"flash_detection.swf?flashContentURL=LNT_content.html&altContentURL=LNT_alternate.html&contentVersion=6&contentMajorRevision=0&contentMinorRevision=0&allowFlashAutoInstall=false\" />
<param name=\"quality\" value=\"low\" />
<embed src=\"flash_detection.swf?flashContentURL=LNT_content.html&altContentURL=LNT_alternate.html&contentVersion=6&contentMajorRevision=0&contentMinorRevision=0&allowFlashAutoInstall=false\" quality=\"low\" pluginspage=\"http://www.macromedia.com/shockwave/download/index.cgi?P1_Prod_Version=ShockwaveFlash\" type=\"application/x-shockwave-flash\" width=\"80\" height=\"80\" />
</object>
</body>
</html>"
))))))

(publish :path "/lifenet/LNT_content.html"
  :content-type "text/html"
  :function
  (lambda (req ent)
    (with-http-response (req ent)
      (with-http-body (req ent)
       (html (:princ
"<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\">
<html xmlns=\"http://www.w3.org/1999/xhtml\" xml:lang=\"en\" lang=\"en\">
<head>
<meta http-equiv=\"Content-Type\" content=\"text/html; charset=iso-8859-1\" />
<title>LNT</title>
</head>
<body bgcolor=\"#ffffff\">
<!--url's used in the movie-->
<!--text used in the movie-->
<object classid=\"clsid:d27cdb6e-ae6d-11cf-96b8-444553540000\" codebase=\"http://download.macromedia.com/pub/shockwave/cabs/flash/swflash.cab#version=6,0,0,0\" width=\"100%\" height=\"100%\" id=\"LNT\" align=\"middle\">
<param name=\"allowScriptAccess\" value=\"sameDomain\" />
<param name=\"movie\" value=\"lifenet/LNT.swf\" />
<param name=\"quality\" value=\"best\" />
<param name=\"bgcolor\" value=\"#ffffff\" />
<embed src=\"lifenet/LNT.swf\" quality=\"best\" bgcolor=\"#ffffff\" width=\"100%\" height=\"100%\" name=\"LNT\" align=\"middle\" allowScriptAccess=\"sameDomain\" type=\"application/x-shockwave-flash\" pluginspage=\"http://www.macromedia.com/go/getflashplayer\" />
</object>
</body>
</html>"
))))))

(publish :path "/lifenet/LNT_alternate.html"
  :content-type "text/html"
  :function
  (lambda (req ent)
    (with-http-response (req ent)
      (with-http-body (req ent)
       (html (:princ 
"<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\">
<html xmlns=\"http://www.w3.org/1999/xhtml\" xml:lang=\"en\" lang=\"en\">
<head>
<title>Macromedia Flash Player Not Current</title>
</head>
<body bgcolor=\"#ffffff\">
<br />
<blockquote>
<h3><font color=\"#999999\" face=\"Arial, Helvetica, sans-serif\">You don't have the latest version of Macromedia Flash Player.</font></h3>
<font face=\"Arial, Helvetica, sans-serif\">
<table width=\"60%\">
  <tr>
    <td colspan=\"2\">
      <p />
      This web site makes use of <a href=\"http://www.macromedia.com/software/flash/\" target=\"_new\">Macromedia<sup><font size=\"-1\">&reg;</font></sup>Flash<sup><font size=\"-2\">TM</font></sup></a> software. You have an old version of Macromedia Flash Player that cannot play the content we've created.
    </td>
  </tr>
  <tr>
    <td>
      <a href=\"http://www.macromedia.com/go/getflashplayer\"><img src=\"alternate.gif\" border=\"0\" width=\"88\" height=\"31\" /></a>
    </td>
    <td>
      <p />
      Why not download and install the latest version now? It will only take a moment.
    </td>
  </tr>
</table>
</font>
<p />
<font face=\"Verdana, Arial, Helvetica, sans-serif\" SIZE=\"-2\">Macromedia and Flash are trademarks of Macromedia, Inc.</font>
</blockquote>
</body>
</html>"
))))))

(defun symbol->string (sym)
  (format nil "~A" sym))

(defun number? (x)
  (typep x 'number))

(defmacro with-cgi-vars ((req) &body body)
   `(let ((cgi-vars (request-query ,req)))
       ,@body))

(defmacro get-cgi-var (cgi-var)
   ;; Get the value of cgi variable with name CGI-VAR.
   ;; Must be called within the scope of WITH-CGI-VARS.
   `(let ((pair (assoc (symbol->string ,cgi-var) cgi-vars :test #'equal)))
       (when pair
         (cdr pair))))

(publish :path "/lifenet/lifenetbegin"
  :content-type "text/plain"
  :function
  (lambda (req ent)
    (with-http-response (req ent)
      (with-http-body (req ent)
        (with-cgi-vars (req)
         (let ((initial-game (kb-get-random-goal)))
          (html
           (:princ "<op>")
           (:princ (format nil "<p id=\"~A\" time=\"~A\" truth=\"~A\">~A</p>"
                       (first initial-game) 0 1 (kb-get-prop-by-id (first initial-game))))
           (:princ (format nil "<p id=\"~A\" time=\"~A\" truth=\"~A\">~A</p>"
                       (second initial-game) 2 1 (kb-get-prop-by-id (second initial-game))))
           (:princ "</op>"))))))))

(publish :path "/lifenet/lifenetsearch"
  :content-type "text/plain"
  :function
  (lambda (req ent)
    (with-http-response (req ent)
      (with-http-body (req ent)
        (with-cgi-vars (req)
          (let ((query (get-cgi-var "query")))
            (cond (query
                   (html (:princ "<props>"))
                   (loop for result in (kb-search-props (simplify query)) do
                         (html
                          (:princ (format nil "<prop id=\"~A\">~A</prop>"
                                     result (kb-get-prop-by-id result)))))
                   (html (:princ "</props>"))))))))))

(publish :path "/lifenet/lifenetupdate"
  :content-type "text/plain"
  :function
  (lambda (req ent)
    (with-http-response (req ent)
      (with-http-body (req ent)
        (with-cgi-vars (req)
          (let* ((updates (get-rule-updates (get-cgi-var "update")))
                  (completion (kb-suggest-completions updates)))
           (kb-update updates)
           (html (:princ "<op>"))
           (loop for result in completion do
                  (html
                   (:princ (format nil "<p id=\"~A\" time=\"~A\" truth=\"~A\">~A</p>"
                               (first result) (second result) (third result) (kb-get-prop-by-id (first result))))))
           (html (:princ "</op>"))))))))

(publish :path "/lifenet/lifenetquery"
  :content-type "text/plain"
  :function
  (lambda (req ent)
    (with-http-response (req ent)
      (with-http-body (req ent)
        (with-cgi-vars (req)
           (print (get-cgi-var "query"))
          (let* ((updates (get-props-from-client (get-cgi-var "query")))
                  (completion (kb-suggest-next updates)))
           (html (:princ "<op>"))
           (loop for result in completion do
                  (html
                   (:princ (format nil "<p id=\"~A\" time=\"~A\" truth=\"~A\">~A</p>"
                               (first result) (second result) (third result) (kb-get-prop-by-id (first result))))))
           (html (:princ "</op>"))))))))


;;; for visualize lifenet system


(defun kb-suggest-completions-2 (props)
  ;; Given a list of propositions supplied as (ID TIME TRUTH) elements, suggest what other
  ;; propositions might be true, returned as (ID TIME TRUTH) elements sorted from highest
  ;; to lowest probability.
  (let ((results nil))
   (loop for prop in props
         for id = (first prop)
         for time = (second prop)
         for truth = (third prop)
         if (= truth 1)
         do
         (setf results (append results
                         (mapcar (lambda (x)
                                   (list (first x) (second x) (- time 1) 1))
                           (kb-get-links-loose id 1 0))))
         (setf results (append results
                         (mapcar (lambda (x)
                                   (list (first x) (second x) time 1))
                           (kb-get-links-loose id 1 1))))
         (setf results (append results 
                         (mapcar (lambda (x)
                                   (list (first x) (second x) (+ time 1) 1)) 
                           (kb-get-links-loose id 0 1)))))
   (remove-duplicates (mapcar #'rest (sort results 
                                       (lambda (x y) (> (first x) (first y)))))
     :test #'equal)))

(defun get-props-from-client-2 (xml)
   (let ((xml-list (parse-xml xml)))
    (loop for elt in (cdar xml-list)
            for id = (read-from-string (third (car elt)))
            for time = (read-from-string (fifth (car elt)))
            for truth = (read-from-string (seventh (car elt)))
            collect (list id time truth))))

(publish :path "/lifenet/lifenet_search_props"
  :content-type "text/plain"
  :function
  (lambda (req ent)
    (with-http-response (req ent)
      (with-http-body (req ent)
        (with-cgi-vars (req)
          (let ((query (get-cgi-var "expr")))
            (cond (query
                   (html (:princ "<props>"))
                   (loop for result in (kb-search-props (simplify query)) do
                         (html
                          (:princ (format nil "<prop id=\"~A\">~A</prop>"
                                     result (kb-get-prop-by-id result)))))
                   (html (:princ "</props>"))))))))))

(publish :path "/lifenet/lifenet_extend_story"
  :content-type "text/plain"
  :function
  (lambda (req ent)
    (with-http-response (req ent)
      (with-http-body (req ent)
        (with-cgi-vars (req)
          (let* ((updates (get-props-from-client-2 (get-cgi-var "evidence")))
                 (completion (kb-suggest-completions-2 updates)))
           (print completion)
           (html (:princ "<op>"))
           (loop for result in completion do
                  (html
                   (:princ (format nil "<p id=\"~A\" time=\"~A\" truth=\"~A\">~A</p>"
                               (first result) (second result) (third result)
                               (kb-get-prop-by-id (first result))))))
           (html (:princ "</op>"))))))))
 

;;; INTERFACE TO STORYNET


;;;~~~~~~~~~~~~~~~~  HTTP-XML Interface to StoryNet  ~~~~~~~~~~~~~~~~

(publish-file
 :path "/storynet/StoryNet.swf"
 :file (merge-pathnames cl-user::*storynet-root* (make-pathname :name "StoryNet" :type "swf")))

(publish-file
 :path "/storynet/flash_detection.swf"
 :file (merge-pathnames cl-user::*lifenet-root* (make-pathname :name "flash_detection" :type "swf")))

(publish :path "/storynet/"
  :content-type "text/html"
  :function
  (lambda (req ent)
    (with-http-response (req ent)
      (with-http-body (req ent)
       (html (:princ 
"<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\">
<html xmlns=\"http://www.w3.org/1999/xhtml\" xml:lang=\"en\" lang=\"en\">
<head>
<meta http-equiv=\"refresh\" content=\"3;url=StoryNet_alternate.html\" />
<title>StoryNet</title>
</head>
<body>
<object classid=\"clsid:d27cdb6e-ae6d-11cf-96b8-444553540000\" codebase=\"http://download.macromedia.com/pub/shockwave/cabs/flash/swflash.cab#version=4,0,0,0\" width=\"80\" height=\"80\">
<param name=\"movie\" value=\"flash_detection.swf?flashContentURL=StoryNet_content.html&altContentURL=StoryNet_alternate.html&contentVersion=6&contentMajorRevision=0&contentMinorRevision=0&allowFlashAutoInstall=false\" />
<param name=\"quality\" value=\"low\" />
<embed src=\"flash_detection.swf?flashContentURL=StoryNet_content.html&altContentURL=StoryNet_alternate.html&contentVersion=6&contentMajorRevision=0&contentMinorRevision=0&allowFlashAutoInstall=false\" quality=\"low\" pluginspage=\"http://www.macromedia.com/shockwave/download/index.cgi?P1_Prod_Version=ShockwaveFlash\" type=\"application/x-shockwave-flash\" width=\"80\" height=\"80\" />
</object>
</body>
</html>"
))))))

(publish :path "/storynet/StoryNet_content.html"
  :content-type "text/html"
  :function
  (lambda (req ent)
    (with-http-response (req ent)
      (with-http-body (req ent)
       (html (:princ
"<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\">
<html xmlns=\"http://www.w3.org/1999/xhtml\" xml:lang=\"en\" lang=\"en\">
<head>
<meta http-equiv=\"Content-Type\" content=\"text/html; charset=iso-8859-1\" />
<title>StoryNet</title>
</head>
<body bgcolor=\"#C0F0B9\">
<!--url's used in the movie-->
<!--text used in the movie-->
<object classid=\"clsid:d27cdb6e-ae6d-11cf-96b8-444553540000\" codebase=\"http://download.macromedia.com/pub/shockwave/cabs/flash/swflash.cab#version=6,0,0,0\" width=\"100%\" height=\"100%\" id=\"StoryNet\" align=\"middle\">
<param name=\"allowScriptAccess\" value=\"sameDomain\" />
<param name=\"movie\" value=\"StoryNet.swf\" />
<param name=\"quality\" value=\"best\" />
<param name=\"bgcolor\" value=\"#c0f0b9\" />
<embed src=\"StoryNet.swf\" quality=\"best\" bgcolor=\"#ffffff\" width=\"100%\" height=\"100%\" name=\"StoryNet\" align=\"middle\" allowScriptAccess=\"sameDomain\" type=\"application/x-shockwave-flash\" pluginspage=\"http://www.macromedia.com/go/getflashplayer\" />
</object>
</body>
</html>"
))))))

(publish :path "/storynet/StoryNet_alternate.html"
  :content-type "text/html"
  :function
  (lambda (req ent)
    (with-http-response (req ent)
      (with-http-body (req ent)
       (html (:princ 
"<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\">
<html xmlns=\"http://www.w3.org/1999/xhtml\" xml:lang=\"en\" lang=\"en\">
<head>
<title>Macromedia Flash Player Not Current</title>
</head>
<body bgcolor=\"#ffffff\">
<br />
<blockquote>
<h3><font color=\"#999999\" face=\"Arial, Helvetica, sans-serif\">You don't have the latest version of Macromedia Flash Player.</font></h3>
<font face=\"Arial, Helvetica, sans-serif\">
<table width=\"60%\">
  <tr>
    <td colspan=\"2\">
      <p />
      This web site makes use of <a href=\"http://www.macromedia.com/software/flash/\" target=\"_new\">Macromedia<sup><font size=\"-1\">&reg;</font></sup>Flash<sup><font size=\"-2\">TM</font></sup></a> software. You have an old version of Macromedia Flash Player that cannot play the content we've created.
    </td>
  </tr>
  <tr>
    <td>
      <a href=\"http://www.macromedia.com/go/getflashplayer\"><img src=\"alternate.gif\" border=\"0\" width=\"88\" height=\"31\" /></a>
    </td>
    <td>
      <p />      Why not download and install the latest version now? It will only take a moment.
    </td>
  </tr>
</table>
</font>
<p />
<font face=\"Verdana, Arial, Helvetica, sans-serif\" SIZE=\"-2\">Macromedia and Flash are trademarks of Macromedia, Inc.</font>
</blockquote>
</body>
</html>"
))))))

